#include<margo.h>
#include<apex.h>
#include"common.h"

int main(int argc, char **argv)
{
    FILE *fd;
    margo_instance_id mid;
    char listen_addr_str[128];
    char server_address[128]; 
    hg_addr_t serv_addr;
    hg_id_t put_id, term_id;
    hg_handle_t puth, termh;
    hg_size_t rdma_size;
    bulk_gdim_t in;
    margo_request req;
    obj_descriptor odsc = {0};
    double *data;
    int i, nsteps;
    hg_return_t hret;

    if(argc != 2) {
        fprintf(stderr, "don't forget step count\n");
        return 1;
    }

    apex_init("client", 0, 1);

    nsteps = atoi(argv[1]);

    fd = fopen("conf", "r");
    fscanf(fd, "%s\n%s", listen_addr_str, server_address);

    mid = margo_init(listen_addr_str, MARGO_SERVER_MODE, 0, 0);
    margo_addr_lookup(mid, server_address, &serv_addr);
    put_id = MARGO_REGISTER(mid, "put_rpc", bulk_gdim_t, bulk_out_t, NULL);
    term_id = MARGO_REGISTER(mid, "term_rpc", void, void, NULL);
    margo_registered_disable_response(mid, term_id, HG_TRUE);

    data = malloc(sizeof(*data));
    odsc.size = sizeof(*data);
    odsc.bb.num_dims = 1;
    odsc.bb.lb.c[0] = odsc.bb.ub.c[0] = 0;
    in.odsc.size = sizeof(odsc);
    in.odsc.raw_odsc = (char *)(&odsc);
    rdma_size = odsc.size;

    sleep(5);

    for(i = 0; i < nsteps; i++) {
        APEX_NAME_TIMER_START(1, "do_put");
        hret = margo_bulk_create(mid, 1, (void **)&data, &rdma_size,
                          HG_BULK_READ_ONLY, &in.handle);
        if(hret != HG_SUCCESS) printf("bulk_create failed %i\n", hret); 
        margo_create(mid, serv_addr, put_id, &puth);
        if(hret != HG_SUCCESS) printf("create failed %i\n", hret);
        margo_iforward(puth, &in, &req);
        if(hret != HG_SUCCESS) printf("iforward failed %i\n", hret);
        APEX_TIMER_STOP(1);
        APEX_NAME_TIMER_START(2, "put_wait");
        margo_wait(req);
        APEX_TIMER_STOP(2);
        margo_bulk_free(in.handle);
        margo_destroy(puth);

    }

    margo_create(mid, serv_addr, term_id, &termh);
    margo_forward(termh, 0);
    margo_addr_free(mid, serv_addr);
    margo_destroy(termh);

    margo_finalize(mid);

    apex_finalize();

    free(data);

    return 0;
}
