#include <margo.h>
#include <mercury.h>
#include <mercury_bulk.h>
#include <mercury_proc_string.h>
#include <mercury_macros.h>

#include <apex.h>

#define APEX_FUNC_TIMER_START(fn)                                              \
    apex_profiler_handle profiler0 = apex_start(APEX_FUNCTION_ADDRESS, &fn);
#define APEX_NAME_TIMER_START(num, name)                                       \
    apex_profiler_handle profiler##num = apex_start(APEX_NAME_STRING, name);
#define APEX_TIMER_STOP(num) apex_stop(profiler##num);

#define ALIGN_ADDR_QUAD_BYTES(a)                                               \
    unsigned long _a = (unsigned long)(a);                                     \
    _a = (_a + 7) & ~7;                                                        \
    (a) = (void *)_a;


struct coord {
    uint64_t c[10];
};

struct bbox {
    int num_dims;
    struct coord lb, ub;
};

enum storage_type { row_major, column_major };

typedef struct {
    char name[150];

    enum storage_type st;
    uint32_t flags;

    char owner[128];
    unsigned int version;

    /* Global bounding box descriptor. */
    struct bbox bb;

    /* Size of one element of a data object. */
    size_t size;
} obj_descriptor;

struct list_head {
    struct list_head *prev, *head;
};

struct global_dimension {
    int ndim;
    struct coord sizes;
};

struct obj_data {
    struct list_head obj_entry;

    obj_descriptor obj_desc;

    struct global_dimension gdim;

    void *data; /* Aligned pointer */

    /* Reference to the parent object; used only for sub-objects. */
    struct obj_data *obj_ref;

    /* Count how many references are to this data object. */
    int refcnt;

    /* Flag to mark if we should free this data object. */
    unsigned int f_free : 1;
};

uint64_t bbox_volume(struct bbox *bb)
{
    uint64_t n = 1;
    int ndims = bb->num_dims;
    int i;

    for(i = 0; i < ndims; i++) {
        n = n * (bb->ub.c[i] - bb->lb.c[i] + 1);
    }
    return n;
}


typedef struct {
    size_t size;
    size_t gdim_size;
    char *raw_odsc;
    char *raw_gdim;
} odsc_hdr_with_gdim;

static inline hg_return_t hg_proc_odsc_hdr_with_gdim(hg_proc_t proc, void *arg)
{
    hg_return_t ret;
    odsc_hdr_with_gdim *in = (odsc_hdr_with_gdim *)arg;
    ret = hg_proc_hg_size_t(proc, &in->size);
    ret = hg_proc_hg_size_t(proc, &in->gdim_size);
    if(ret != HG_SUCCESS)
        return ret;
    if(in->size) {
        switch(hg_proc_get_op(proc)) {
        case HG_ENCODE:
            ret = hg_proc_raw(proc, in->raw_odsc, in->size);
            if(ret != HG_SUCCESS)
                return ret;
            ret = hg_proc_raw(proc, in->raw_gdim, in->gdim_size);
            if(ret != HG_SUCCESS)
                return ret;
            break;
        case HG_DECODE:
            in->raw_odsc = (char *)malloc(in->size);
            ret = hg_proc_raw(proc, in->raw_odsc, in->size);
            if(ret != HG_SUCCESS)
                return ret;
            in->raw_gdim = (char *)malloc(in->gdim_size);
            ret = hg_proc_raw(proc, in->raw_gdim, in->gdim_size);
            if(ret != HG_SUCCESS)
                return ret;
            break;
        case HG_FREE:
            free(in->raw_odsc);
            free(in->raw_gdim);
            break;
        default:
            break;
        }
    }
    return HG_SUCCESS;
}

MERCURY_GEN_PROC(bulk_gdim_t, ((odsc_hdr_with_gdim)(odsc))((hg_bulk_t)(handle)))
MERCURY_GEN_PROC(bulk_out_t, ((int32_t)(ret)))

