#include <margo.h>
#include <mercury.h>
#include <mercury_macros.h>

#include "common.h"

#include <apex.h>

static void put_rpc(hg_handle_t handle)
{
    margo_instance_id mid = margo_hg_handle_get_instance(handle);
    const struct hg_info *info = margo_get_info(handle);
    bulk_gdim_t in;
    bulk_out_t out;
    hg_bulk_t bulk_handle;
    obj_descriptor in_odsc;
    hg_addr_t owner_addr;
    size_t owner_addr_size = 128;
    struct obj_data od = {0};
    hg_size_t size;

    apex_init("server", 0, 1);

    APEX_NAME_TIMER_START(1, "put_setup");
    margo_get_input(handle, &in);
    memcpy(&in_odsc, in.odsc.raw_odsc, sizeof(in_odsc));
    margo_addr_self(mid, &owner_addr);
    margo_addr_to_string(mid, in_odsc.owner, &owner_addr_size, owner_addr);
    margo_addr_free(mid, owner_addr);

    size = in_odsc.size * bbox_volume(&in_odsc.bb);
    od.data = malloc(size);
    ALIGN_ADDR_QUAD_BYTES(od.data);
    APEX_TIMER_STOP(1);
    
    APEX_NAME_TIMER_START(2, "create_handle");
    margo_bulk_create(mid, 1, (void **)&od.data, &size, 
                      HG_BULK_WRITE_ONLY, &bulk_handle);
    APEX_TIMER_STOP(2);

    APEX_NAME_TIMER_START(3, "transfer");
    margo_bulk_transfer(mid, HG_BULK_PULL, info->addr, in.handle, 0,
                       bulk_handle, 0, size);
    APEX_TIMER_STOP(3);

    APEX_NAME_TIMER_START(4, "respond");
    out.ret = 0;
    margo_respond(handle, &out);
    APEX_TIMER_STOP(4);
    margo_bulk_free(bulk_handle);
    margo_free_input(handle, &in);
    margo_destroy(handle);
}
DEFINE_MARGO_RPC_HANDLER(put_rpc)

static void term_rpc(hg_handle_t handle)
{
    margo_instance_id mid = margo_hg_handle_get_instance(handle);

    margo_destroy(handle);
    margo_finalize(mid);
}
DEFINE_MARGO_RPC_HANDLER(term_rpc)

int main(int argc, char **argv)
{
    char *listen_addr_str;
    margo_instance_id mid;
    hg_id_t put_id, term_id;
    hg_addr_t my_addr = HG_ADDR_NULL;
    char *my_addr_str = NULL;
    hg_size_t my_addr_size = 0;
    FILE *fd;

    if(argc != 2) {
        fprintf(stderr, "don't forget listen string\n");
        return 1;
    }

    listen_addr_str = argv[1];

    mid = margo_init(listen_addr_str, MARGO_SERVER_MODE, 1, 4);
    put_id = MARGO_REGISTER(mid, "put_rpc", bulk_gdim_t, bulk_out_t, put_rpc);
    term_id = MARGO_REGISTER(mid, "term_rpc", void, void, term_rpc);
    margo_registered_disable_response(mid, term_id, HG_TRUE);

    margo_addr_self(mid, &my_addr);
    margo_addr_to_string(mid, NULL, &my_addr_size, my_addr);
    my_addr_str = malloc(my_addr_size);
    margo_addr_to_string(mid, my_addr_str, &my_addr_size, my_addr);
    fd = fopen("conf", "w");
    fprintf(fd, "%s\n%s\n", listen_addr_str, my_addr_str);
    fclose(fd); 
    free(my_addr_str);

    margo_wait_for_finalize(mid);

    apex_finalize();

    return 0;
}
